"""
Plot correlations between user characteristics on the one hand and activity efforts and scenario ratings on the other hand.

Author: Nele Albers
Date: January 2022
"""


import matplotlib.pylab as plt
import numpy as np
import os
import pandas as pd
import seaborn as sns
import scipy.stats


# current directory
curr_directory = os.path.abspath(os.path.dirname(__file__))


if __name__ == "__main__":
    
    # Load file with all coded data
    df = pd.read_csv("Data/all_coded_data.csv")
    # lists are saved as multi-line strings
    df['themes'] = [i.split(' ') for i in df['themes'].values]
    df['themes'] = [[int(j.strip("[]\n.")) for j in i if j.strip("[]\n.") in ["0", "1"]] for i in df['themes']]
    
    
    codes_subset = ["PA_Identity", "Quitting_Self_Identity", 
                    "Nonsmoker_Self_Identity",
                    "TTM_PA", "Weekly Exercise",
                    "Smoking frequency", "Highest education level completed",
                    "Household Size", "age", "Pers_Agreeableness",
                    "Pers_Conscientiousness", "Pers_ES", "Pers_Extraversion",
                    "Pers_OE"]
    y_labels = ["PA Identity", "Quitter Self-Identity", "Non-Smoker Self-Identity", 
                "TTM-Stage PA",
                "Weekly Exercise", "Smoking Frequency",
                "Education Level", "Household Size", "Age", "Agreeableness",
                "Conscientiousness", "Emotional Stability", "Extraversion",
                "Openness to Exp."]
    
    resp_types = [["activity_experience1", "activity_experience2", "activity_experience3",
                   "activity_experience4", "activity_experience5"],
                  ["pa_program"],
                  ["hrs_plan_pa", "hrs_plan_smoking", "hrs_reflection_smoking", 
                   "hrs_reflection_pa"],
                  ["hrs_help_button_pa", "hrs_help_button_smoking"],
                  ["failing_pa_goals"], 
                  ["motivational_messages"],
                   ["tell_se", "discussion_so"], 
                  ["gp_start", "gp_relapse"]]
    x_labels = ["Effort put into Pre-\nparatory Activities", 
                "Follow PA Program", 
                "Regular HRS\nPlanning/Reflection",
                "Help Button for HRS", 
                "Discuss Repeated\nFailure PA", 
                "Receive Motivational\nMessages",
                 "Involve SE", 
                "Involve GP"]
    
    corrs = np.zeros((len(codes_subset), len(resp_types)))
    
    # Compute correlation coefficients
    for r_idx in range(len(resp_types)):
        
        df_temp = df[df["response_type"].isin(resp_types[r_idx])]
        
        for v_idx in range(len(codes_subset)):
            
            df_temp_v = df_temp.dropna(subset = [codes_subset[v_idx]])
            
            corrs[v_idx, r_idx] = scipy.stats.spearmanr(df_temp_v["rating"].to_list(), df_temp_v[codes_subset[v_idx]].to_list())[0]
     
    colors = sns.color_palette("Blues", as_cmap=True)
    ax = sns.heatmap(abs(corrs), linewidth=0.5, cmap=colors, 
                     xticklabels = x_labels, yticklabels = y_labels)
    
    # Create labels within heatmap.
    cutoff = 0.2  # 0.3 is a moderate correlation, 0.1 a small one
    for r_idx in range(len(resp_types)):
        for v_idx in range(len(codes_subset)):
            if abs(corrs[v_idx, r_idx]) >= cutoff:
                text = ax.text(r_idx + 0.5, v_idx + 0.5, round(corrs[v_idx, r_idx], 2),
                               ha="center", va="center", color="w", size = 5)
                
    # Create vertical line to separate activity experience
    ax.axvline(x=1, linewidth=5, color="w")
    
    # Add some white space around plot so that all axis labels are there
    plt.subplots_adjust(left=0.32)
    plt.subplots_adjust(bottom=0.40)
    
    plt.savefig("Figures/correlations_user_characteristics.jpg", dpi = 1500,
                bbox_inches='tight', pad_inches=0)
    