"""
The purpose of this file is to compute the agreement between the two coders.

Required files:
    - Data/All_codes.xlsx
    - Data/Workbook_Codes_Second_Coder.xlsx
    - Data/Workbook_Post_Questionnaire_Anonym_Preprocessed_Coded.xlsx
    - Data/Workbook_Activity_Experiences_Anonym_Preprocessed_Coded.xlsx

Output file: 
    - calculate_agreement_between_coders_output.txt.

Author: Nele Albers
Date: March 2022
"""


import numpy as np
import pandas as pd
from sklearn.metrics import cohen_kappa_score


def brennan_prediger(codes1, codes2, q):
    """
    Calculates the Brennan-Prediger coefficient.

    Parameters
    ----------
    codes1 : list
        Codes given by first coder with possible codes 0, ...., q-1.
    codes2 : list
        Codes given by second coder with possible codes 0, ..., q-1.
    q : int
        Number of possible codes/response categories. Cannot be 1.

    Returns
    -------
    float
        Brennan-Prediger coefficient.

    """

    p_a = sum([1 for i in range(len(codes1)) if codes1[i] == codes2[i]])/len(codes1)
    p_e = 1/q  # q is the number of response categories

    return (p_a - p_e)/(1 - p_e)


def get_higher_level_codes_from_lower_level(code_level_low, num_levellow_per_levelhigh):
    """
    Compute the higher-level code for a given input code based on max, i.e.
    the higher-level code is 1 if at least one of its lower-level codes is 1.
    For example, we can compute the codes at level 2 based on codes at level 3.

    Parameters
    ----------
    code_level_low : list
        Binary code for a response with 0/1 for each code.
    num_levellow_per_levelhigh : list
        Number of lower-level codes for each high-level code.

    Returns
    -------
    code_level_high: list
        Binary code for the response with 0/1 for each higher-level code.

    """

    num_codes_level_high = len(num_levellow_per_levelhigh)
    code_level_high = np.zeros(num_codes_level_high)

    for code_high in range(num_codes_level_high):
        code_level_high[code_high] = max(code_level_low[sum(num_levellow_per_levelhigh[0:code_high]):sum(num_levellow_per_levelhigh[0:code_high + 1])])

    return code_level_high


def calc_agreement_coefficients(matching_codes, codes_binary_sec_coder, 
                                all_codes_list):
    """
    Calculate and print several agreement coefficients.

    Parameters
    ----------
    matching_codes : list of lists
        Contains for each response one list with 0/1 for each code based on coding by the first coder.
    codes_binary_sec_coder : list of lists
        Contains for each response one list with 0/1 for each code based on coding by the second coder.
    all_codes_list : list
        List of all codes that have been used for the coding.

    Returns
    -------
    None.

    """

    # First flatten lists
    matching_codes_flat = [i for sublist in matching_codes for i in sublist]
    codes_binary_sec_coder_flat = [i for sublist in codes_binary_sec_coder for i in sublist]

    # Now compute cohen's kappa and print result
    overall_cohens_kappa = cohen_kappa_score(matching_codes_flat, codes_binary_sec_coder_flat)
    print("\nCohen's Kappa:", round(overall_cohens_kappa, 2))
    with open(output_file_name, 'a') as f:
        f.write("Cohen's Kappa: " + str(round(overall_cohens_kappa, 2)) + "\n")

    # Calculate overall Brennan-prediger
    overall_b_p = brennan_prediger(matching_codes_flat, codes_binary_sec_coder_flat, 2)
    print("Brennan-Prediger Kappa:", round(overall_b_p, 2))
    with open(output_file_name, 'a') as f:
        f.write("Brennan-Predigar Kappa: " + str(round(overall_b_p, 2)) + "\n")


def get_codes_post_q(num_codes_per_data_type_postq,
                     df_first_coder_postq,
                     all_codes_list,
                     additional_offsets_postq):
    """Get codes from first coder for post-questionnaire free-text responses."""

    codes_binary_first_coder_postq = []  # Codes at level 3
    all_user_responses_with_na = []  # for indexing with responses from second coder

    for data_type in range(len(num_codes_per_data_type_postq)):
        code_binary_curr_data_type = []

        for row in range(len(df_first_coder_postq)):  # For each response
            code_binary_curr_row = np.zeros(len(all_codes_list))

            for code_idx in range(num_codes_per_data_type_postq[data_type]):
                col_idx_higher_level = int(2 + code_idx * 3 + (data_type + 1) * 2 + sum(num_codes_per_data_type_postq[0:data_type]) * 3 + sum(additional_offsets_postq[0:data_type]))
                col_idx_lower_level = col_idx_higher_level + 1

                user_response = ""
                if code_idx == 0:
                    user_response =  df_first_coder_postq.iloc[row, col_idx_higher_level - 2]
                    all_user_responses_with_na.append(user_response)

                if not pd.isna(user_response):
                    code_higher_level = df_first_coder_postq.iloc[row, col_idx_higher_level]
                    code_lower_level = df_first_coder_postq.iloc[row, col_idx_lower_level]

                    if not code_lower_level == "\\":  # no code
                        code_found = code_higher_level + " - " + code_lower_level
                        code_found_idx = all_codes_list.index(code_found)
                        code_binary_curr_row[code_found_idx] = 1

            code_binary_curr_data_type.append(code_binary_curr_row)

        codes_binary_first_coder_postq.append(code_binary_curr_data_type)
        

    return codes_binary_first_coder_postq, all_user_responses_with_na


def get_codes_session(num_codes_per_data_type_session,
                      all_codes_list,
                      df_first_coder_session):

    """Get codes from first coder given to session data."""

    codes_binary_first_coder_session = []
    all_user_responses_with_na = []

    for data_type in range(len(num_codes_per_data_type_session)):
        code_binary_curr_data_type = []

        for row in range(len(df_first_coder_session)):  # For each response
            code_binary_curr_row = np.zeros(len(all_codes_list))

            for code_idx in range(num_codes_per_data_type_session[data_type]):
                col_idx_higher_level = int(1 + code_idx * 3 + (data_type + 1) * 2 + sum(num_codes_per_data_type_session[0:data_type]) * 3)
                col_idx_lower_level = col_idx_higher_level + 1

                user_response = ""
                if code_idx == 0:
                    user_response =  df_first_coder_session.iloc[row, col_idx_higher_level - 2]
                    all_user_responses_with_na.append(user_response)

                if not pd.isna(user_response):
                    code_higher_level = df_first_coder_session.iloc[row, col_idx_higher_level]
                    code_lower_level = df_first_coder_session.iloc[row, col_idx_lower_level]

                    if not code_lower_level == "\\":  # no code
                        code_found = code_higher_level + " - " + code_lower_level
                        code_found_idx = all_codes_list.index(code_found)
                        code_binary_curr_row[code_found_idx] = 1

            code_binary_curr_data_type.append(code_binary_curr_row)

        codes_binary_first_coder_session.append(code_binary_curr_data_type)


    return codes_binary_first_coder_session, all_user_responses_with_na


if __name__ == "__main__":

    # To save output
    output_file_name = "calculate_agreement_between_coders_output.txt"

    all_codes_df = pd.read_excel("Data/All_codes.xlsx")

    # Load coding of second coder
    df_second_coder = pd.read_excel("Data/Workbook_Codes_Second_Coder.xlsx")
    num_sec_coder = len(df_second_coder)
    max_num_codes_sec_coder = 6


    # Load coding of first coder for post-questionnaire data
    df_first_coder_postq = pd.read_excel("Data/Workbook_Post_Questionnaire_Anonym_Preprocessed_Coded.xlsx")
    # Only the first 500 rows contain data
    df_first_coder_postq = df_first_coder_postq.head(500)
    # Max. number of codes for each data type/question in the post-questionnaire
    num_codes_per_data_type_postq = [5, 4, 4, 6, 6, 6, 6, 5, 4, 4, 4, 4, 4, 4, 6, 4]
    # Some additional columns for some data types: video-rating columns.
    additional_offsets_postq = [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0]


    # Load coding of first coder for session data
    df_first_coder_session = pd.read_excel("Data/Workbook_Activity_Experiences_Anonym_Preprocessed_Coded.xlsx")
    num_codes_per_data_type_session = [5, 5, 7, 5, 5, 5, 5, 5]


    # Create list of all codes at the 3 levels
    all_codes_list = []  # all codes at level 3
    all_codes_list_level2 = []  # All codes at level 2
    all_codes_list_level1 = ["Other party", "Behavior", "User", "Environment"]
    num_level3_per_level2 = []  # Number of sub-levels per level 2
    num_level2_per_level1 = [4, 7, 3, 1]

    for col_idx, col in enumerate(all_codes_df.columns.to_list()):
        all_codes_list_level2.append(col)
        num_level3_per_level2.append(0)
        for code in all_codes_df[col].to_list():
            if not pd.isna(code):
                all_codes_list.append(col + " - " + code)
                num_level3_per_level2[col_idx] += 1

    # Extract codes of first coder for post-questionnaire data
    print("... Extract codes of first coder for post-questionnaire data")
    codes_binary_first_coder_postq, all_user_responses_with_na = get_codes_post_q(num_codes_per_data_type_postq,
                                                                                  df_first_coder_postq,
                                                                                  all_codes_list,
                                                                                  additional_offsets_postq)


    # Extract codes of first coder for session data
    print("... Extract codes of first coder for session data")
    codes_binary_first_coder_session, all_user_responses_with_na_session = get_codes_session(num_codes_per_data_type_session,
                                                                                             all_codes_list,
                                                                                             df_first_coder_session)
    # Concatenate all user responses
    all_user_responses_with_na += all_user_responses_with_na_session


    # Get codes of first coder corresponding to the responses the second coder
    # coded
    responses_coded_by_second_coder = df_second_coder["Response"].to_list()
    codes_binary_first_coder_postq_flat = [i for sublist in codes_binary_first_coder_postq for i in sublist]
    matching_codes = []  # codes by first coder for responses coded by second coder
    
    for resp_2 in responses_coded_by_second_coder:
        idx_1 = all_user_responses_with_na.index(resp_2)
        matching_codes.append(codes_binary_first_coder_postq_flat[idx_1])


    # Extract codes of second coder
    print("... Extract codes of second coder")
    codes_binary_sec_coder = []
    for row in range(num_sec_coder):
        code_binary_curr = np.zeros(len(all_codes_list))

        for code in range(max_num_codes_sec_coder):
            col_idx_higher_level = int(code * 3 + 2)
            col_idx_lower_level = col_idx_higher_level + 1
            code_higher_level = df_second_coder.iloc[row, col_idx_higher_level]
            code_lower_level = df_second_coder.iloc[row, col_idx_lower_level]

            if not code_lower_level == "\\":  # no code
                code_found = code_higher_level + " - " + code_lower_level
                code_found_idx = all_codes_list.index(code_found)
                code_binary_curr[code_found_idx] = 1

        codes_binary_sec_coder.append(code_binary_curr)


    # Calculate agreement coefficients for each coding level.
    print("\n\nAgreement coefficients:")
    print("\n**Level 3**")
    with open(output_file_name, 'w') as f:
        f.write("\n***Level 3\n")

    calc_agreement_coefficients(matching_codes, 
                                codes_binary_sec_coder, 
                                all_codes_list)

    print("\n**Level 2 (based on max.)**")
    with open(output_file_name, 'a') as f:
        f.write("\n***Level 2\n")

    # Compute codes at level 2 based on max (i.e., value for a second-level code is 1 if at least one of the corresponding third-level codes is 1).
    matching_codes_level2 = [get_higher_level_codes_from_lower_level(i, num_level3_per_level2) for i in matching_codes]
    codes_binary_sec_coder_level2 = [get_higher_level_codes_from_lower_level(i, num_level3_per_level2) for i in codes_binary_sec_coder]

    calc_agreement_coefficients(matching_codes_level2, 
                                codes_binary_sec_coder_level2, 
                                all_codes_list_level2)

    print("\n**Level 1 (based on max.)**")
    with open(output_file_name, 'a') as f:
        f.write("\n***Level 1\n")

    # Compute codes at level 1 based on max.
    matching_codes_level1 = [get_higher_level_codes_from_lower_level(i, num_level2_per_level1) for i in matching_codes_level2]
    codes_binary_sec_coder_level1 = [get_higher_level_codes_from_lower_level(i, num_level2_per_level1) for i in codes_binary_sec_coder_level2]

    calc_agreement_coefficients(matching_codes_level1, 
                                codes_binary_sec_coder_level1, 
                                all_codes_list_level1)
