"""
The purpose of this file is to process the coded data to facilitate analysis.
We also compute the Cronbach's alpha values for indices we create 
(e.g., physical activity identity).

Required files:
    - Data/pre_questionnaire_anonym.csv
    - Data/prescreening_questionnaire_anonym.csv
    - Data/prescreening_prolific_numeric_anonym.csv
    - Data/All_codes.xlsx
    - Data/Workbook_Post_Questionnaire_Anonym_Preprocessed_Coded.xlsx
    - Data/Workbook_Activity_Experiences_Anonym_Preprocessed_Coded.xlsx

Output files: 
    - process_after_coding_for_analysis_output.txt
    - Correlations/Data/all_coded_data.csv
    - Correlations/Bayesian_analysis/Data/all_coded_data.csv
    - Code_frequencies/Data/all_coded_data.csv
    - Scenario_Ratings_and_Participant_Characteristics/Data/all_coded_data.csv

Author: Nele Albers
Date: June 2022
"""


import numpy as np
import pandas as pd


def cronbach_alpha(df):
    """
    Calculate cronbach's alpha.
    Code from https://towardsdatascience.com/cronbachs-alpha-theory-and-application-in-python-d2915dd63586.
    """

    # 1. Transform the df into a correlation matrix
    df_corr = df.corr()

    # 2.1 Calculate N
    # The number of variables equals the number of columns in the df
    N = df.shape[1]

    # 2.2 Calculate R
    # For this, we'll loop through the columns and append every
    # relevant correlation to an array called "r_s". Then, we'll
    # calculate the mean of "r_s"
    rs = np.array([])
    for i, col in enumerate(df_corr.columns):
        sum_ = df_corr[col][i+1:].values
        rs = np.append(sum_, rs)
    mean_r = np.mean(rs)

    # 3. Use the formula to calculate Cronbach's Alpha 
    cronbach_alpha = (N * mean_r) / (1 + (N - 1) * mean_r)

    return cronbach_alpha


def get_char_pre_questionnaire(df, output_file_name):
    """Get user characteristics from pre-questionnaire."""

    df_preq_q = pd.read_csv("Data/pre_questionnaire_anonym.csv")
    df_preq_q = df_preq_q[["TTM_PA", "rand_id",
                           "Personality_1", "Personality_2", "Personality_3",
                           "Personality_4", "Personality_5", "Personality_6",
                           "Personality_7", "Personality_8", "Personality_9",
                           "Personality_10", "PA_Identity_1", "PA_Identity_2",
                           "PA_Identity_3", "PA_Identity_4", "PA_Identity_5",
                           "PA_Identity_6", "PA_Identity_7", "PA_Identity_8",
                           "PA_Identity_9", "Quitting_Self_Identity_1",
                           "Quitting_Self_Identity_2", "Quitting_Self_Identity_3",
                           "Nonsmoker_Self_Identity_1", 
                           "Nonsmoker_Self_Identity_2", "Nonsmoker_Self_Identity_3",
                           "Quit_Before_24h"]]

    df = pd.merge(df, df_preq_q, on = "rand_id")

    df['Pers_Extraversion'] = 0.5 * (df['Personality_1']- (df['Personality_6'] - 8)) 
    df['Pers_Agreeableness'] = 0.5 * (df['Personality_7'] - (df['Personality_2'] - 8)) 
    df['Pers_Conscientiousness'] = 0.5 * (df['Personality_3'] - (df['Personality_8'] - 8)) 
    df['Pers_ES'] = 0.5 * (df['Personality_9'] - (df['Personality_4'] - 8)) 
    df['Pers_OE'] = 0.5 * (df['Personality_5'] - (df['Personality_10'] - 8)) 
    df["PA_Identity"] = 1/9 * (df["PA_Identity_1"] + df["PA_Identity_2"] + df["PA_Identity_3"] + df["PA_Identity_4"] + df["PA_Identity_5"] + df["PA_Identity_6"] + df["PA_Identity_8"] + df["PA_Identity_9"] + df["PA_Identity_7"])  
    df["Quitting_Self_Identity"] = 1/3 * (df["Quitting_Self_Identity_1"] + df["Quitting_Self_Identity_2"] + df["Quitting_Self_Identity_3"]) 
    df["Nonsmoker_Self_Identity"] = 1/3 * (df["Nonsmoker_Self_Identity_1"] + df["Nonsmoker_Self_Identity_2"] + df["Nonsmoker_Self_Identity_3"]) 

    # Single entry per user for computing cronbach's alpha
    df_single = df[df["response_type"] == "activity_experience1"]

    # Compute cronbach's alpha
    df_sub = df_single[["Quitting_Self_Identity_1", "Quitting_Self_Identity_2", 
                        "Quitting_Self_Identity_3"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha quitter self-identity:", round(alpha, 2))
    with open(output_file_name, 'w') as f:
        f.write("Cronbach's alpha quitter self-identity: " + str(round(alpha, 2)) + "\n")

    df_sub = df_single[["Nonsmoker_Self_Identity_1", "Nonsmoker_Self_Identity_2", 
                        "Nonsmoker_Self_Identity_3"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha non-smoker self-identity:", round(alpha, 2))
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha non-smoker self-identity: " + str(round(alpha, 2)) + "\n")

    df_sub = df_single[["PA_Identity_1", "PA_Identity_2",
                        "PA_Identity_3", "PA_Identity_4", "PA_Identity_5",
                        "PA_Identity_6", "PA_Identity_7", "PA_Identity_8",
                        "PA_Identity_9"]]
    alpha = cronbach_alpha(df_sub)
    print("Cronbach's alpha physical activity identity:", round(alpha, 2))
    with open(output_file_name, 'a') as f:
        f.write("Cronbach's alpha physical activity identity: " + str(round(alpha, 2)) + "\n")

    df = df.drop(columns=["Personality_1", "Personality_2", "Personality_3",
                          "Personality_4", "Personality_5", "Personality_6",
                          "Personality_7", "Personality_8", "Personality_9",
                          "Personality_10", "PA_Identity_1", "PA_Identity_2",
                          "PA_Identity_3", "PA_Identity_4", "PA_Identity_5",
                          "PA_Identity_6", "PA_Identity_7", "PA_Identity_8",
                          "PA_Identity_9", "Quitting_Self_Identity_1",
                          "Quitting_Self_Identity_2", "Quitting_Self_Identity_3",
                          "Nonsmoker_Self_Identity_1", 
                          "Nonsmoker_Self_Identity_2", 
                          "Nonsmoker_Self_Identity_3"])

    # Let's reverse the scale for the TTM-stage for PA such that a higher
    # value denotes a higher stage of change.
    df["TTM_PA"] = abs(df["TTM_PA"] - 6)

    return df


def get_char_pre_screening(df):
    """Get user characteristics from prescreening questionnaire."""

    # 1 stands for wanting to quit within next 30 days, 2 for within 6 months.
    df_prescreening_q = pd.read_csv("Data/prescreening_questionnaire_anonym.csv")[["Quit_Smoking", "rand_id"]]
    df = pd.merge(df, df_prescreening_q, on = "rand_id", how = 'left')

    # People are only in preparation stage if they also had at least one quit
    # attempt of at least 24h in the last year. This info is in the pre-questionnaire.
    df_preq_q = pd.read_csv("Data/pre_questionnaire_anonym.csv")[["Num_Quit_24h_Last_Year", "rand_id"]]
    df= pd.merge(df, df_preq_q, on = "rand_id")

    # 1 stands for preparation phase, 2 for contemplation phase
    df["TTM_Smoking"] = [1 if df.iloc[i]["Quit_Smoking"] == 1 and df.iloc[i]["Num_Quit_24h_Last_Year"] >= 2 else 2 for i in range(len(df))]

    df = df.drop(columns = ["Quit_Smoking"])

    return df


def get_char_prolific(df):
    """Get user characteristics from Prolific profiles."""
    
    df_presc_prol = pd.read_csv("Data/prescreening_prolific_numeric_anonym.csv")[["age_bin", "Gender identity", 
                                                                                  "rand_id", "Smoking frequency", 
                                                                                  "Weekly Exercise", "Household Size", 
                                                                                  "Highest education level completed"]]

    df = pd.merge(df, df_presc_prol, on="rand_id")

    return df


def get_single_dataframe_with_all_responses(df_first_coder_postq,
                                            codes_binary_first_coder_postq,
                                            df_first_coder_session,
                                            codes_binary_first_coder_session):
    """Create a single dataframe with all free-text responses."""

    # Now we want to have a single dataframe with all responses
    response_types = []
    responses = []
    ratings = []
    codes = []
    users = []

    # First the data from the post-questionnaire
    # Names of columns with free-text responses in post-questionnaire coding
    # workbook
    cols_postq = ["Experience with last assigned activity", 
                  "Barriers to doing activities",
                  "Motivators for doing activities",
                  "E1 - Have a Reflection on Difficult Situations w.r.t. PA in last Week with VC on Sundays (https://www.youtube.com/watch?v=FXC7oZLLzvQ&ab_channel=NA)",
                  "E2 - Reflection with Virtual Coach on Difficult Smoking-related Situations in the Evening (https://www.youtube.com/watch?v=Raq7Qyvh45A&ab_channel=NA)",
                  'E3 - Tell SE About Quit Attempt (https://www.youtube.com/watch?v=wDxMd-vwJgM&ab_channel=NA)',
                  'E4 - Receive Motivational Messages from Virtual Coach (https://www.youtube.com/watch?v=IGKRjhK40Cs&ab_channel=NA)',
                  'E5 - Consult GP at Start of Quitting Process (https://www.youtube.com/watch?v=lDTeHgBA71w&ab_channel=NA)',
                  'E6 - Plan for Difficult Situations w.r.t. Smoking with Virtual Coach in the Morning (https://www.youtube.com/watch?v=zpWy28vtSzM&ab_channel=NA)',
                  'E7 - Use a Help Button for HRS for Smoking (https://www.youtube.com/watch?v=sLBF2BuMv80&ab_channel=NeleAlbers)',
                  'E8 - Follow PA Program Besides Quitting Smoking (https://www.youtube.com/watch?v=_0lmuAuJbfU&ab_channel=NA)',
                  'E9 - Use a Help Button for HRS w.r.t. PA (https://www.youtube.com/watch?v=HZLtG4LfvIs&ab_channel=NeleAlbers)',
                  'E10 - Create Weekly Plan for Difficult Situations w.r.t. PA with Virtual Coach on Sundays (https://www.youtube.com/watch?v=cA78pzMoFuo&ab_channel=NeleAlbers)',
                  'E11 - Consult GP in Case of a Relapse w.r.t. Smoking (https://www.youtube.com/watch?v=Rl-xdeRexUw&ab_channel=NA)',
                  'E12 - Contact Virtual Coach when Repeatedly Failing PA Goals (https://www.youtube.com/watch?v=STO8tEGpWXI&ab_channel=NA)',
                  'E13 - Have a Discussion with Social Environment About How they Can Support Quit Attempt (https://www.youtube.com/watch?v=ZbopPvhQJWw&ab_channel=NA)',
                  ]

    ratings_postq = ["Effort_1", "", "", "R1_1", "R2_1", "R3_1", "R4_1", "R5_1",
                     "R6_1", "R7_1", "R8_1", "R9_1", "R10_1", "R11_1", "R12_1", 
                     "R13_1"]

    cols_resp_types = ["activity_experience5",
                       "barriers",
                       "motivators",
                       "hrs_reflection_pa",
                       "hrs_reflection_smoking",
                       "tell_se",
                       "motivational_messages",
                       "gp_start",
                       "hrs_plan_smoking",
                       "hrs_help_button_smoking",
                       "pa_program",
                       "hrs_help_button_pa",
                       "hrs_plan_pa",
                       "gp_relapse",
                       "failing_pa_goals",
                       "discussion_so"]


    for col in range(len(cols_postq)):
        for user_idx in range(len(df_first_coder_postq)):

            # not everybody saw each video and 4 times persons wrote in the answer that they could not open the video
            if not pd.isnull(df_first_coder_postq.iloc[user_idx][cols_postq[col]]) and not "no results found when I pasted to YouTube" in df_first_coder_postq.iloc[user_idx][cols_postq[col]] and not "t see it" in df_first_coder_postq.iloc[user_idx][cols_postq[col]] and not "Same issue as before" in df_first_coder_postq.iloc[user_idx][cols_postq[col]]:
                response_types.append(cols_resp_types[col])
                responses.append(df_first_coder_postq.iloc[user_idx][cols_postq[col]])

                if not cols_postq[col] in ["Barriers to doing activities", "Motivators for doing activities"]:
                    ratings.append(df_first_coder_postq.iloc[user_idx][ratings_postq[col]])
                else:
                    ratings.append(np.nan)

                users.append(df_first_coder_postq.iloc[user_idx]['rand_id'])
                codes.append(codes_binary_first_coder_postq[col][user_idx])


    # Now let's add the session data
    cols_session = ["activity_experience1", 
                    "activity_experience2",
                    "activity_experience3",
                    "activity_experience4", 
                    "activity_experience_mod1",
                    "activity_experience_mod2", 
                    "activity_experience_mod3",
                    "activity_experience_mod4"]


    for col in range(len(cols_session)):
        for user_idx in range(len(df_first_coder_session)):

            if not pd.isnull(df_first_coder_session.iloc[user_idx][cols_session[col]]): # not everybody answered everything
                response_types.append(cols_session[col])
                responses.append(df_first_coder_session.iloc[user_idx][cols_session[col]])

                if not "mod" in cols_session[col]:  # there are no new ratings for exp. modifications
                    rew_list = str(df_first_coder_session.iloc[user_idx]["reward_list"])
                    rew_list = rew_list.split('|')
                    ratings.append(rew_list[col])
                else:
                    ratings.append(np.nan)

                users.append(df_first_coder_session.iloc[user_idx]['rand_id'])
                codes.append(codes_binary_first_coder_session[col][user_idx])


    # Create dataframe
    df = pd.DataFrame(response_types, columns=["response_type"])
    df['response'] = responses
    df['rating'] = ratings
    df['rand_id'] = users
    df['themes'] = codes


    return df


def get_codes_post_q(num_codes_per_data_type_postq,
                     df_first_coder_postq,
                     all_codes_list,
                     additional_offsets_postq):
    """Get codes from first coder given to post-questionnaire data."""

    codes_binary_first_coder_postq = []  # Codes at level 3

    for data_type in range(len(num_codes_per_data_type_postq)):
        code_binary_curr_data_type = []

        for row in range(len(df_first_coder_postq)):  # For each response
            code_binary_curr_row = np.zeros(len(all_codes_list))

            for code_idx in range(num_codes_per_data_type_postq[data_type]):

                col_idx_higher_level = int(2 + code_idx * 3 + (data_type + 1) * 2 + sum(num_codes_per_data_type_postq[0:data_type]) * 3 + sum(additional_offsets_postq[0:data_type]))
                col_idx_lower_level = col_idx_higher_level + 1

                user_response = ""
                if code_idx == 0:
                    user_response =  df_first_coder_postq.iloc[row, col_idx_higher_level - 2]

                if not pd.isna(user_response):

                    code_higher_level = df_first_coder_postq.iloc[row, col_idx_higher_level]
                    code_lower_level = df_first_coder_postq.iloc[row, col_idx_lower_level]

                    if not code_lower_level == "\\":  # no code

                        code_found = code_higher_level + " - " + code_lower_level
                        code_found_idx = all_codes_list.index(code_found)
                        code_binary_curr_row[code_found_idx] = 1

            code_binary_curr_data_type.append(code_binary_curr_row)

        codes_binary_first_coder_postq.append(code_binary_curr_data_type)


    return codes_binary_first_coder_postq


def get_codes_session(num_codes_per_data_type_session,
                      all_codes_list,
                      df_first_coder_session):
    """Get codes from first coder given to session data."""

    codes_binary_first_coder_session = []

    for data_type in range(len(num_codes_per_data_type_session)):
        code_binary_curr_data_type = []

        for row in range(len(df_first_coder_session)):  # For each response
            code_binary_curr_row = np.zeros(len(all_codes_list))

            for code_idx in range(num_codes_per_data_type_session[data_type]):

                col_idx_higher_level = int(1 + code_idx * 3 + (data_type + 1) * 2 + sum(num_codes_per_data_type_session[0:data_type]) * 3)
                col_idx_lower_level = col_idx_higher_level + 1

                user_response = ""
                if code_idx == 0:
                    user_response =  df_first_coder_session.iloc[row, col_idx_higher_level - 2]

                if not pd.isna(user_response):

                    code_higher_level = df_first_coder_session.iloc[row, col_idx_higher_level]
                    code_lower_level = df_first_coder_session.iloc[row, col_idx_lower_level]

                    if not code_lower_level == "\\":  # no code

                        code_found = code_higher_level + " - " + code_lower_level
                        code_found_idx = all_codes_list.index(code_found)
                        code_binary_curr_row[code_found_idx] = 1

            code_binary_curr_data_type.append(code_binary_curr_row)

        codes_binary_first_coder_session.append(code_binary_curr_data_type)


    return codes_binary_first_coder_session


def get_all_codes():
    """Get all codes from coding scheme."""

    all_codes_df = pd.read_excel("Data/All_codes.xlsx")

    all_codes_list = []  # all codes at level 3

    for col_idx, col in enumerate(all_codes_df.columns.to_list()):
        for code in all_codes_df[col].to_list():
            if not pd.isna(code):
                all_codes_list.append(col + " - " + code)

    return all_codes_list


if __name__ == "__main__":


    # To save output (i.e., Cronbach's alpha values)
    output_file_name = "process_after_coding_for_analysis_output.txt"


    # Load coding of first coder for post-questionnaire data
    df_first_coder_postq = pd.read_excel("Data/Workbook_Post_Questionnaire_Anonym_Preprocessed_Coded.xlsx")
    # Only the first 500 rows contain data
    df_first_coder_postq = df_first_coder_postq.head(500)
    # Max. number of codes for each data type/question in the post-questionnaire
    num_codes_per_data_type_postq = [5, 4, 4, 6, 6, 6, 6, 5, 4, 4, 4, 4, 4, 4, 6, 4]
    # Some additional columns for some data types: video-rating columns.
    additional_offsets_postq = [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0]


    # Load coding of first coder for session data
    df_first_coder_session = pd.read_excel("Data/Workbook_Activity_Experiences_Anonym_Preprocessed_Coded.xlsx")
    num_codes_per_data_type_session = [5, 5, 7, 5, 5, 5, 5, 5]


    # Get a list with all codes
    all_codes_list = get_all_codes()


    # Extract codes of first coder for post-questionnaire data
    print("... Extract codes of first coder for post-questionnaire data")
    codes_binary_first_coder_postq = get_codes_post_q(num_codes_per_data_type_postq,
                                                      df_first_coder_postq,
                                                      all_codes_list,
                                                      additional_offsets_postq)


    # Extract codes of first coder for session data
    print("... Extract codes of first coder for session data")
    codes_binary_first_coder_session = get_codes_session(num_codes_per_data_type_session,
                                                         all_codes_list,
                                                         df_first_coder_session)


    # Get a single dataframe with all responses (from sessions and 
    # post-questionnaire)
    df = get_single_dataframe_with_all_responses(df_first_coder_postq,
                                                 codes_binary_first_coder_postq,
                                                 df_first_coder_session,
                                                 codes_binary_first_coder_session)


    # Let's add user variables
    print("... Extract data on participants from prescreening questionnaire")
    # First, variables from the prescreening questionnaire.
    df = get_char_pre_screening(df)

    print("... Extract data on participants from pre-questionnaire")
    # Second, variables from the pre-questionnaire (e.g., personality)
    df = get_char_pre_questionnaire(df, output_file_name)

    print("... Extract data on participants from Prolific profiles")
    # Third, let's add some variables from Prolific (e.g., age)
    df = get_char_prolific(df)


    # Let's also give each user a simple number as ID so that we can talk
    # about the user in the paper.
    all_rand_ids = list(sorted(set(df["rand_id"].to_list()))) # unique rand IDs
    df["Participant_Num"] = [all_rand_ids.index(df.iloc[i]["rand_id"]) + 1 for i in range(len(df))]


    # Add a separate binary column for each code
    num_codes = len(df.iloc[0]["themes"])
    for t in range(num_codes):
        df["Code_" + str(t)] = [df.iloc[i]["themes"][t] for i in range(len(df))]


    # Save the dataframe to wherever we need it
    df.to_csv("Correlations/Data/all_coded_data.csv")
    df.to_csv("Correlations/Bayesian_analysis/Data/all_coded_data.csv")
    df.to_csv("Code_frequencies/Data/all_coded_data.csv")
    df.to_csv("Scenario_Ratings_and_Participant_Characteristics/Data/all_coded_data.csv")
